@@ -31,6 +31,7 @@ import time
 import struct
 import socket
 import fcntl
+import traceback
 import debconf
 import apt_pkg
 from apt.package import Package
@@ -88,23 +89,30 @@ class DebconfFetchProgress(FetchProgress):
 class DebconfInstallProgress(InstallProgress):
     """An object that reports apt's installation progress using debconf."""
 
-    def __init__(self, db, title, info, error):
+    def __init__(self, db, title, info, error=None):
         InstallProgress.__init__(self)
         self.db = db
         self.title = title
         self.info = info
         self.error_template = error
         self.started = False
+        # InstallProgress uses a non-blocking status fd; our run()
+        # implementation doesn't need that, and in fact we spin unless the
+        # fd is blocking.
+        flags = fcntl.fcntl(self.statusfd.fileno(), fcntl.F_GETFL)
+        fcntl.fcntl(self.statusfd.fileno(), fcntl.F_SETFL,
+                    flags & ~os.O_NONBLOCK)
 
     def startUpdate(self):
         self.db.progress('START', 0, 100, self.title)
         self.started = True
 
     def error(self, pkg, errormsg):
-        self.db.subst(self.error_template, 'PACKAGE', pkg)
-        self.db.subst(self.error_template, 'MESSAGE', errormsg)
-        self.db.input('critical', self.error_template)
-        self.db.go()
+        if self.error_template is not None:
+            self.db.subst(self.error_template, 'PACKAGE', pkg)
+            self.db.subst(self.error_template, 'MESSAGE', errormsg)
+            self.db.input('critical', self.error_template)
+            self.db.go()
 
     def statusChange(self, pkg, percent, status):
         self.percent = percent
@@ -117,52 +125,98 @@ class DebconfInstallProgress(InstallProgress):
         # TODO cjwatson 2006-02-28: InstallProgress.updateInterface doesn't
         # give us a handy way to spot when percentages/statuses change and
         # aren't pmerror/pmconffile, so we have to reimplement it here.
-        if self.statusfd != None:
-            try:
-                while not self.read.endswith("\n"):
-                    self.read += os.read(self.statusfd.fileno(),1)
-            except OSError, (err,errstr):
-                # resource temporarily unavailable is ignored
-                if err != errno.EAGAIN:
-                    print errstr
-            if self.read.endswith("\n"):
-                s = self.read
-                (status, pkg, percent, status_str) = s.split(":", 3)
-                if status == "pmerror":
-                    self.error(pkg, status_str)
-                elif status == "pmconffile":
-                    # we get a string like this:
-                    # 'current-conffile' 'new-conffile' useredited distedited
-                    match = re.compile("\s*\'(.*)\'\s*\'(.*)\'.*").match(status_str)
-                    if match:
-                        self.conffile(match.group(1), match.group(2))
-                else:
-                    self.statusChange(pkg, float(percent), status_str.strip())
-                self.read = ""
+        if self.statusfd is None:
+            return False
+        try:
+            while not self.read.endswith("\n"):
+                r = os.read(self.statusfd.fileno(),1)
+                if not r:
+                    return False
+                self.read += r
+        except OSError, (err,errstr):
+            print errstr
+        if self.read.endswith("\n"):
+            s = self.read
+            (status, pkg, percent, status_str) = s.split(":", 3)
+            if status == "pmerror":
+                self.error(pkg, status_str)
+            elif status == "pmconffile":
+                # we get a string like this:
+                # 'current-conffile' 'new-conffile' useredited distedited
+                match = re.compile("\s*\'(.*)\'\s*\'(.*)\'.*").match(status_str)
+                if match:
+                    self.conffile(match.group(1), match.group(2))
+            else:
+                self.statusChange(pkg, float(percent), status_str.strip())
+            self.read = ""
+        return True
 
     def run(self, pm):
-        pid = self.fork()
-        if pid == 0:
+        # Create a subprocess to deal with turning apt status messages into
+        # debconf protocol messages.
+        child_pid = self.fork()
+        if child_pid == 0:
             # child
+            os.close(self.writefd)
+            try:
+                while self.updateInterface():
+                    pass
+            except (KeyboardInterrupt, SystemExit):
+                pass # we're going to exit anyway
+            except:
+                traceback.print_exc(file=sys.stderr)
+            os._exit(0)
 
-            # Redirect stdout to stderr to avoid it interfering with our
-            # debconf protocol stream.
-            os.dup2(2, 1)
+        self.statusfd.close()
 
-            # Make sure all packages are installed non-interactively. We
-            # don't have enough passthrough magic here to deal with any
-            # debconf questions they might ask.
-            os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
-            if 'DEBIAN_HAS_FRONTEND' in os.environ:
-                del os.environ['DEBIAN_HAS_FRONTEND']
-            if 'DEBCONF_USE_CDEBCONF' in os.environ:
-                # Probably not a good idea to use this in /target too ...
-                del os.environ['DEBCONF_USE_CDEBCONF']
+        # Redirect stdout to stderr to avoid it interfering with our
+        # debconf protocol stream.
+        saved_stdout = os.dup(1)
+        os.dup2(2, 1)
 
+        # Make sure all packages are installed non-interactively. We
+        # don't have enough passthrough magic here to deal with any
+        # debconf questions they might ask.
+        saved_environ_keys = ('DEBIAN_FRONTEND', 'DEBIAN_HAS_FRONTEND',
+                              'DEBCONF_USE_CDEBCONF')
+        saved_environ = {}
+        for key in saved_environ_keys:
+            if key in os.environ:
+                saved_environ[key] = os.environ[key]
+        os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
+        if 'DEBIAN_HAS_FRONTEND' in os.environ:
+            del os.environ['DEBIAN_HAS_FRONTEND']
+        if 'DEBCONF_USE_CDEBCONF' in os.environ:
+            # Probably not a good idea to use this in /target too ...
+            del os.environ['DEBCONF_USE_CDEBCONF']
+
+        res = pm.ResultFailed
+        try:
             res = pm.DoInstall(self.writefd)
-            sys.exit(res)
-        self.child_pid = pid
-        res = self.waitChild()
+        finally:
+            # Reap the status-to-debconf subprocess.
+            os.close(self.writefd)
+            while True:
+                try:
+                    (pid, status) = os.waitpid(child_pid, 0)
+                    if pid != child_pid:
+                        break
+                    if os.WIFEXITED(status) or os.WIFSIGNALED(status):
+                        break
+                except OSError:
+                    break
+
+            # Put back stdout.
+            os.dup2(saved_stdout, 1)
+            os.close(saved_stdout)
+
+            # Put back the environment.
+            for key in saved_environ_keys:
+                if key in saved_environ:
+                    os.environ[key] = saved_environ[key]
+                elif key in os.environ:
+                    del os.environ[key]
+
         return res
 
     def finishUpdate(self):
@@ -186,6 +240,7 @@ class Install:
 
         apt_pkg.InitConfig()
         apt_pkg.Config.Set("Dir", "/target")
+        apt_pkg.Config.Set("Dir::State::status", "/target/var/lib/dpkg/status")
         apt_pkg.Config.Set("APT::GPGV::TrustedKeyring",
                            "/target/etc/apt/trusted.gpg")
         apt_pkg.Config.Set("Acquire::gpgv::Options::",
@@ -196,6 +251,23 @@ class Install:
         apt_pkg.Config.Clear("DPkg::Pre-Install-Pkgs")
         apt_pkg.InitSystem()
 
+    def excepthook(self, exctype, excvalue, exctb):
+        """Crash handler. Dump the traceback to a file so that it can be
+        read by the caller."""
+
+        if (issubclass(exctype, KeyboardInterrupt) or
+            issubclass(exctype, SystemExit)):
+            return
+
+        tbtext = ''.join(traceback.format_exception(exctype, excvalue, exctb))
+        print >>sys.stderr, "Exception during installation:"
+        print >>sys.stderr, tbtext
+        tbfile = open('/var/lib/ubiquity/install.trace', 'w')
+        print >>tbfile, tbtext
+        tbfile.close()
+
+        sys.exit(1)
+
     def run(self):
         """Run the install stage: copy everything to the target system, then
         configure it as necessary."""
@@ -208,93 +280,93 @@ class Install:
                 return False
 
         self.db.progress('SET', 1)
-        self.db.progress('REGION', 1, 78)
+        self.db.progress('REGION', 1, 75)
         if not self.copy_all():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 78)
+        self.db.progress('SET', 75)
         self.db.progress('INFO', 'ubiquity/install/cleanup')
         if self.source == '/source':
             if not self.umount_source():
                 self.db.progress('STOP')
                 return False
 
-        self.db.progress('SET', 79)
-        self.db.progress('REGION', 79, 80)
+        self.db.progress('SET', 76)
+        self.db.progress('REGION', 76, 77)
         if not self.run_target_config_hooks():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 80)
-        self.db.progress('REGION', 80, 81)
+        self.db.progress('SET', 77)
+        self.db.progress('REGION', 77, 78)
         self.db.progress('INFO', 'ubiquity/install/locales')
         if not self.configure_locales():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 81)
-        self.db.progress('REGION', 81, 82)
+        self.db.progress('SET', 78)
+        self.db.progress('REGION', 78, 79)
         self.db.progress('INFO', 'ubiquity/install/network')
         if not self.configure_network():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 82)
-        self.db.progress('REGION', 82, 83)
+        self.db.progress('SET', 79)
+        self.db.progress('REGION', 79, 80)
         self.db.progress('INFO', 'ubiquity/install/apt')
         if not self.configure_apt():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 83)
-        self.db.progress('REGION', 83, 87)
+        self.db.progress('SET', 80)
+        self.db.progress('REGION', 80, 84)
         # Ignore failures from language pack installation.
         self.install_language_packs()
 
-        self.db.progress('SET', 87)
-        self.db.progress('REGION', 87, 88)
+        self.db.progress('SET', 84)
+        self.db.progress('REGION', 84, 85)
         self.db.progress('INFO', 'ubiquity/install/timezone')
         if not self.configure_timezone():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 88)
-        self.db.progress('REGION', 88, 90)
+        self.db.progress('SET', 85)
+        self.db.progress('REGION', 85, 87)
         self.db.progress('INFO', 'ubiquity/install/keyboard')
         if not self.configure_keyboard():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 90)
-        self.db.progress('REGION', 90, 91)
+        self.db.progress('SET', 87)
+        self.db.progress('REGION', 87, 88)
         self.db.progress('INFO', 'ubiquity/install/user')
         if not self.configure_user():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 91)
-        self.db.progress('REGION', 91, 95)
+        self.db.progress('SET', 88)
+        self.db.progress('REGION', 88, 92)
         self.db.progress('INFO', 'ubiquity/install/hardware')
         if not self.configure_hardware():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 95)
-        self.db.progress('REGION', 95, 96)
+        self.db.progress('SET', 92)
+        self.db.progress('REGION', 92, 93)
         if not self.remove_unusable_kernels():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 96)
-        self.db.progress('REGION', 96, 97)
+        self.db.progress('SET', 93)
+        self.db.progress('REGION', 93, 94)
         self.db.progress('INFO', 'ubiquity/install/bootloader')
         if not self.configure_bootloader():
             self.db.progress('STOP')
             return False
 
-        self.db.progress('SET', 97)
-        self.db.progress('REGION', 97, 99)
+        self.db.progress('SET', 94)
+        self.db.progress('REGION', 94, 99)
         self.db.progress('INFO', 'ubiquity/install/removing')
         if not self.remove_extras():
             self.db.progress('STOP')
@@ -670,8 +742,7 @@ class Install:
             self.db, 'ubiquity/langpacks/title', None,
             'ubiquity/langpacks/packages')
         installprogress = DebconfInstallProgress(
-            self.db, 'ubiquity/langpacks/title', 'ubiquity/install/apt_info',
-            'ubiquity/install/apt_error_install')
+            self.db, 'ubiquity/langpacks/title', 'ubiquity/install/apt_info')
 
         for lp in to_install:
             self.mark_install(cache, lp)
@@ -1182,7 +1253,14 @@ class Install:
 
 
 if __name__ == '__main__':
-    if Install().run():
+    if not os.path.exists('/var/lib/ubiquity'):
+        os.makedirs('/var/lib/ubiquity')
+    if os.path.exists('/var/lib/ubiquity/install.trace'):
+        os.unlink('/var/lib/ubiquity/install.trace')
+
+    install = Install()
+    sys.excepthook = install.excepthook
+    if install.run():
         sys.exit(0)
     else:
         sys.exit(1)
