@@ -449,6 +449,10 @@ class Wizard:
 
         disable_swap()
 
+        child = self.embedded.get_child()
+        if child is not None:
+            self.embedded.remove(child)
+
         socket = gtk.Socket()
         socket.show()
         self.embedded.add(socket)
@@ -469,8 +473,13 @@ class Wizard:
         # widget is studied in a different manner depending on object type
         if widget.__class__ == str:
             size = float(self.size[widget.split('/')[2]])
-        else:
+        elif widget.get_active_text() in self.part_devices:
             size = float(self.size[self.part_devices[widget.get_active_text()].split('/')[2]])
+        else:
+            # TODO cjwatson 2006-07-31: Why isn't it in part_devices? This
+            # indicates a deeper problem somewhere, but for now we'll just
+            # try our best to ignore it.
+            return ''
 
         if size > 1024*1024:
             msg = '%.0f Gb' % (size/1024/1024)
@@ -534,10 +543,16 @@ class Wizard:
         ret = dbfilter.run_command(auto_process=True)
         if ret != 0:
             self.installing = False
-            # TODO cjwatson 2006-05-23: figure out why Install crashed
-            raise RuntimeError, ("Install failed with exit code %s; see "
-                                 "/var/log/installer/syslog and "
-                                 "/var/log/syslog" % ret)
+            if os.path.exists('/var/lib/ubiquity/install.trace'):
+                tbfile = open('/var/lib/ubiquity/install.trace')
+                realtb = tbfile.read()
+                tbfile.close()
+                raise RuntimeError, ("Install failed with exit code %s\n%s" %
+                                     (ret, realtb))
+            else:
+                raise RuntimeError, ("Install failed with exit code %s; see "
+                                     "/var/log/installer/syslog and "
+                                     "/var/log/syslog" % ret)
 
         while self.progress_position.depth() != 0:
             self.debconf_progress_stop()
@@ -664,8 +679,7 @@ class Wizard:
             self.hostname.handler_unblock(self.hostname_delete_text_id)
 
         complete = True
-        for name in ('fullname', 'username', 'password', 'verified_password',
-                     'hostname'):
+        for name in ('username', 'password', 'verified_password', 'hostname'):
             if getattr(self, name).get_text() == '':
                 complete = False
         self.allow_go_forward(complete)
@@ -806,11 +820,45 @@ class Wizard:
             self.next.set_label("Install") # TODO i18n
 
 
+    def gparted_crashed(self):
+        """gparted crashed. Ask the user if they want to continue."""
+
+        # TODO cjwatson 2006-07-18: i18n
+        text = ('The advanced partitioner (gparted) crashed. Further '
+                'information may be found in /var/log/installer/syslog, '
+                'or by running gparted directly. Do you want to try the '
+                'advanced partitioner again, return to automatic '
+                'partitioning, or quit this installer?')
+        dialog = gtk.Dialog('GParted crashed', self.live_installer,
+                            gtk.DIALOG_MODAL,
+                            (gtk.STOCK_QUIT, gtk.RESPONSE_CLOSE,
+                             'Automatic partitioning', 1,
+                             'Try again', 2))
+        label = gtk.Label(text)
+        label.set_line_wrap(True)
+        label.set_selectable(True)
+        dialog.vbox.add(label)
+        dialog.show_all()
+        response = dialog.run()
+        dialog.hide()
+        if response == 1:
+            self.steps.set_current_page(self.steps.page_num(self.stepPartDisk))
+        elif response == gtk.RESPONSE_CLOSE:
+            self.current_page = None
+            self.quit()
+        else:
+            self.gparted_loop()
+
+
     def gparted_to_mountpoints(self):
         """Processing gparted to mountpoints step tasks."""
 
         self.gparted_fstype = {}
 
+        if self.gparted_subp is None:
+            self.gparted_crashed()
+            return
+
         try:
             print >>self.gparted_subp.stdin, "apply"
         except IOError:
@@ -818,6 +866,7 @@ class Wizard:
             self.gparted_subp.stdin.close()
             self.gparted_subp.wait()
             self.gparted_subp = None
+            self.gparted_crashed()
             return
 
         # read gparted output of format "- FORMAT /dev/hda2 linux-swap"
@@ -829,6 +878,7 @@ class Wizard:
                 self.gparted_fstype[words[1]] = words[2]
             gparted_reply = \
                 self.gparted_subp.stdout.readline().rstrip('\n')
+        pre_log('info', 'gparted replied: %s' % gparted_reply)
 
         if gparted_reply.startswith('1 '):
             # Cancel
@@ -854,59 +904,67 @@ class Wizard:
             self.part_devices[label] = partition
             self.partition_choices.append(partition)
 
-        # Initialise the mountpoints table.
-        if len(self.mountpoint_widgets) == 0:
-            self.add_mountpoint_table_row()
+        # Reinitialise the mountpoints table.
+        for child in self.mountpoint_table.get_children():
+            if child.get_name() not in ('mountpoint_label', 'size_label',
+                                        'device_label', 'format_label'):
+                self.mountpoint_table.remove(child)
+        self.mountpoint_widgets = []
+        self.size_widgets = []
+        self.partition_widgets = []
+        self.format_widgets = []
 
-            # Try to get some default mountpoint selections.
-            self.size = get_sizes()
-            selection = get_default_partition_selection(
-                self.size, self.gparted_fstype, self.auto_mountpoints)
+        self.add_mountpoint_table_row()
 
-            # Setting a default partition preselection
-            if len(selection.items()) == 0:
-                self.allow_go_forward(False)
-            else:
-                # Setting default preselection values into ComboBox
-                # widgets and setting size values. In addition, next row
-                # is showed if they're validated.
-                for mountpoint, partition in selection.items():
-                    if partition.split('/')[2] not in self.size:
-                        continue
-                    if partition not in self.partition_choices:
-                        # TODO cjwatson 2006-05-27: I don't know why this
-                        # might happen, but it does
-                        # (https://launchpad.net/bugs/46910). Figure out
-                        # why. In the meantime, ignoring this partition is
-                        # better than crashing.
-                        continue
-                    if mountpoint in self.mountpoint_choices:
-                        self.mountpoint_widgets[-1].set_active(
-                            self.mountpoint_choices.index(mountpoint))
-                    else:
-                        self.mountpoint_widgets[-1].child.set_text(mountpoint)
-                    self.size_widgets[-1].set_text(
-                        self.set_size_msg(partition))
-                    self.partition_widgets[-1].set_active(
-                        self.partition_choices.index(partition))
-                    if (mountpoint in ('swap', '/', '/usr', '/var', '/boot') or
-                        partition in self.gparted_fstype):
-                        self.format_widgets[-1].set_active(True)
-                    else:
-                        self.format_widgets[-1].set_active(False)
-                    if partition not in self.gparted_fstype:
-                        self.format_widgets[-1].set_sensitive(True)
-                    if len(get_partitions()) > len(self.partition_widgets):
-                        self.add_mountpoint_table_row()
-                    else:
-                        break
+        # Try to get some default mountpoint selections.
+        self.size = get_sizes()
+        selection = get_default_partition_selection(
+            self.size, self.gparted_fstype, self.auto_mountpoints)
 
-            # We defer connecting up signals until now to avoid the changed
-            # signal firing while we're busy populating the table.
-            for mountpoint in self.mountpoint_widgets:
-                mountpoint.connect("changed", self.on_list_changed)
-            for partition in self.partition_widgets:
-                partition.connect("changed", self.on_list_changed)
+        # Setting a default partition preselection
+        if len(selection.items()) == 0:
+            self.allow_go_forward(False)
+        else:
+            # Setting default preselection values into ComboBox widgets and
+            # setting size values. In addition, the next row is shown if
+            # they're validated.
+            for mountpoint, partition in selection.items():
+                if partition.split('/')[2] not in self.size:
+                    continue
+                if partition not in self.partition_choices:
+                    # TODO cjwatson 2006-05-27: I don't know why this might
+                    # happen, but it does
+                    # (https://launchpad.net/bugs/46910). Figure out why. In
+                    # the meantime, ignoring this partition is better than
+                    # crashing.
+                    continue
+                if mountpoint in self.mountpoint_choices:
+                    self.mountpoint_widgets[-1].set_active(
+                        self.mountpoint_choices.index(mountpoint))
+                else:
+                    self.mountpoint_widgets[-1].child.set_text(mountpoint)
+                self.size_widgets[-1].set_text(
+                    self.set_size_msg(partition))
+                self.partition_widgets[-1].set_active(
+                    self.partition_choices.index(partition))
+                if (mountpoint in ('swap', '/', '/usr', '/var', '/boot') or
+                    partition in self.gparted_fstype):
+                    self.format_widgets[-1].set_active(True)
+                else:
+                    self.format_widgets[-1].set_active(False)
+                if partition not in self.gparted_fstype:
+                    self.format_widgets[-1].set_sensitive(True)
+                if len(get_partitions()) > len(self.partition_widgets):
+                    self.add_mountpoint_table_row()
+                else:
+                    break
+
+        # We defer connecting up signals until now to avoid the changed
+        # signal firing while we're busy populating the table.
+        for mountpoint in self.mountpoint_widgets:
+            mountpoint.connect("changed", self.on_list_changed)
+        for partition in self.partition_widgets:
+            partition.connect("changed", self.on_list_changed)
 
         self.mountpoint_error_reason.hide()
         self.mountpoint_error_image.hide()
@@ -925,7 +983,10 @@ class Wizard:
             mountpoint_value = self.mountpoint_widgets[i].get_active_text()
             partition_value = self.partition_widgets[i].get_active_text()
             if partition_value is not None:
-                partition_id = self.part_devices[partition_value]
+                if partition_value in self.part_devices:
+                    partition_id = self.part_devices[partition_value]
+                else:
+                    partition_id = partition_value
             else:
                 partition_id = None
             format_value = self.format_widgets[i].get_active()
@@ -965,7 +1026,15 @@ class Wizard:
 
         # Processing more validation stuff
         if len(self.mountpoints) > 0:
-            for check in validation.check_mountpoint(self.mountpoints,
+            # Supplement filesystem types from gparted FORMAT instructions
+            # with those detected from the disk.
+            validate_mountpoints = dict(self.mountpoints)
+            validate_filesystems = get_filesystems(self.gparted_fstype)
+            for device, (path, format, fstype) in validate_mountpoints.items():
+                if fstype is None and device in validate_filesystems:
+                    validate_mountpoints[device] = \
+                        (path, format, validate_filesystems[device])
+            for check in validation.check_mountpoint(validate_mountpoints,
                                                      self.size):
                 if check == validation.MOUNTPOINT_NOROOT:
                     error_msg.append(get_string(
@@ -988,6 +1057,22 @@ class Wizard:
                     error_msg.append(get_string(
                         'partman-basicfilesystems/bad_mountpoint',
                         self.locale))
+                elif check == validation.MOUNTPOINT_XFSROOT:
+                    error_msg.append("XFS may not be used on the filesystem "
+                                     "containing /boot. Either use a "
+                                     "different filesystem for / or create a "
+                                     "non-XFS filesystem for /boot.")
+                elif check == validation.MOUNTPOINT_XFSBOOT:
+                    error_msg.append("XFS may not be used on the /boot "
+                                     "filesystem. Use a different filesystem "
+                                     "type for /boot.")
+                elif check == validation.MOUNTPOINT_UNFORMATTED:
+                    error_msg.append("Filesystems used by the system (/, "
+                                     "/boot, /usr, /var) must be reformatted "
+                                     "for use by this installer. Other "
+                                     "filesystems (/home, /media/*, "
+                                     "/usr/local, etc.) may be used without "
+                                     "reformatting.")
 
         # showing warning messages
         self.mountpoint_error_reason.set_text("\n".join(error_msg))
@@ -1477,7 +1562,7 @@ class Wizard:
         return True
 
     def set_keyboard_choices(self, choicemap):
-        self.keyboard_choice_map = choicemap
+        self.keyboard_choice_map = dict(choicemap)
         choices = choicemap.keys()
 
         kbdlayouts = gtk.ListStore(gobject.TYPE_STRING)
@@ -1530,8 +1615,11 @@ class Wizard:
 
 
     def return_to_autopartitioning (self):
-        """Return from the install progress bar to autopartitioning."""
-        if self.installing:
+        """If the install progress bar is up but still at the partitioning
+        stage, then errors can safely return us to autopartitioning.
+        """
+
+        if self.installing and self.current_page is not None:
             # Go back to the autopartitioner and try again.
             # TODO self.previous_partitioning_page
             self.live_installer.show()
@@ -1734,6 +1822,9 @@ class TimezoneMap(object):
             return None
 
     def location_from_point(self, point):
+        if point is None:
+            return None
+
         (longitude, latitude) = point.get_location()
 
         best_location = None
