@@ -184,9 +184,8 @@ class Wizard:
         self.autopartition_buttongroup_texts = {}
         
         self.qtparted_vbox = QVBoxLayout(self.userinterface.qtparted_frame)
-        self.embed = QXEmbed(self.userinterface.qtparted_frame, "embed")
-        self.embed.setProtocol(QXEmbed.XPLAIN)
-        
+        self.embed = None
+        self.mountpoint_table = QGridLayout(self.userinterface.mountpoint_frame, 2, 4, 11, 6)
 
     def excepthook(self, exctype, excvalue, exctb):
         """Crash handler."""
@@ -354,15 +353,16 @@ class Wizard:
                          core_names=['ubiquity/text/%s' % q
                                      for q in self.language_questions])
 
-        self.translate_widget_chidren(parentWidget)
+        self.translate_widget_children(parentWidget)
 
-    def translate_widget_chidren(self, parentWidget=None):
+    def translate_widget_children(self, parentWidget=None):
         if parentWidget == None:
             parentWidget = self.userinterface
 
-        for widget in parentWidget.children():
-            self.translate_widget(widget, self.locale)
-            self.translate_widget_chidren(widget)
+        if parentWidget.children() != None:
+            for widget in parentWidget.children():
+                self.translate_widget(widget, self.locale)
+                self.translate_widget_children(widget)
 
     def translate_widget(self, widget, lang):
 
@@ -432,13 +432,19 @@ class Wizard:
         label_text = label_text.replace("${TOTAL}", str(BREADCRUMB_MAX_STEP))
         self.userinterface.step_label.setText(label_text)
 
-    def gparted_loop(self):
-        """call gparted and embed it into glade interface."""
+    def qtparted_loop(self):
+        """call qtparted and embed it into the interface."""
 
-        pre_log('info', 'gparted_loop()')
+        pre_log('info', 'qtparted_loop()')
 
         disable_swap()
 
+        if self.embed is not None:
+            self.qtparted_vbox.remove(self.embed)
+            del self.embed
+        self.embed = QXEmbed(self.userinterface.qtparted_frame, "embed")
+        self.embed.setProtocol(QXEmbed.XPLAIN)
+
         self.qtparted_subp = subprocess.Popen(
             ['/usr/sbin/qtparted', '--installer'], stdin=subprocess.PIPE, stdout=subprocess.PIPE, close_fds=True)
         qtparted_winid = self.qtparted_subp.stdout.readline().rstrip('\n')
@@ -457,8 +463,13 @@ class Wizard:
         # widget is studied in a different manner depending on object type
         if widget.__class__ == str:
             size = float(self.size[widget.split('/')[2]])
-        else:
+        elif str(widget.currentText()) in self.part_devices:
             size = float(self.size[self.part_devices[str(widget.currentText())].split('/')[2]])
+        else:
+            # TODO cjwatson 2006-07-31: Why isn't it in part_devices? This
+            # indicates a deeper problem somewhere, but for now we'll just
+            # try our best to ignore it.
+            return ''
 
         if size > 1024*1024:
             msg = '%.0f Gb' % (size/1024/1024)
@@ -520,10 +531,16 @@ class Wizard:
         ret = dbfilter.run_command(auto_process=True)
         if ret != 0:
             self.installing = False
-            # TODO cjwatson 2006-05-23: figure out why Install crashed
-            raise RuntimeError, ("Install failed with exit code %s; see "
-                                 "/var/log/installer/syslog and "
-                                 "/var/log/syslog" % ret)
+            if os.path.exists('/var/lib/ubiquity/install.trace'):
+                tbfile = open('/var/lib/ubiquity/install.trace')
+                realtb = tbfile.read()
+                tbfile.close()
+                raise RuntimeError, ("Install failed with exit code %s\n%s" %
+                                     (ret, realtb))
+            else:
+                raise RuntimeError, ("Install failed with exit code %s; see "
+                                     "/var/log/installer/syslog and "
+                                     "/var/log/syslog" % ret)
 
         while self.progress_position.depth() != 0:
             self.debconf_progress_stop()
@@ -607,7 +624,7 @@ class Wizard:
                 self.format_widgets[index].setChecked(False)
             else:
                 partition = self.part_devices[partition_text]
-                if partition in self.gparted_fstype:
+                if partition in self.qtparted_fstype:
                     self.format_widgets[index].setEnabled(False)
                     self.format_widgets[index].setChecked(True)
                 else:
@@ -639,8 +656,7 @@ class Wizard:
             self.userinterface.hostname.blockSignals(False)
 
         complete = True
-        for name in ('fullname', 'username', 'password', 'verified_password',
-                     'hostname'):
+        for name in ('username', 'password', 'verified_password', 'hostname'):
             if getattr(self.userinterface, name).text() == '':
                 complete = False
         self.userinterface.next.setEnabled(complete)
@@ -728,7 +744,7 @@ class Wizard:
             self.process_autopartitioning()
         # Advanced partitioning
         elif step == "stepPartAdvanced":
-            self.gparted_to_mountpoints()
+            self.qtparted_to_mountpoints()
         # Mountpoints
         elif step == "stepPartMountpoints":
             self.mountpoints_to_summary()
@@ -773,7 +789,7 @@ class Wizard:
         # then go to manual partitioning.
         choice = self.get_disk_choice()
         if self.manual_choice is None or choice == self.manual_choice:
-            self.gparted_loop()
+            self.qtparted_loop()
             self.userinterface.widgetStack.raiseWidget(WIDGET_STACK_STEPS["stepPartAdvanced"])
         else:
             self.userinterface.widgetStack.raiseWidget(WIDGET_STACK_STEPS["stepPartAuto"])
@@ -787,7 +803,7 @@ class Wizard:
         # then go to manual partitioning.
         choice = self.get_autopartition_choice()
         if self.manual_choice is None or choice == self.manual_choice:
-            self.gparted_loop()
+            self.qtparted_loop()
             self.userinterface.widgetStack.raiseWidget(WIDGET_STACK_STEPS["stepPartAdvanced"])
         else:
             # TODO cjwatson 2006-01-10: extract mountpoints from partman
@@ -795,10 +811,34 @@ class Wizard:
             installText = get_string("live_installer", self.locale)
             self.userinterface.next.setText(installText)
 
-    def gparted_to_mountpoints(self):
-        """Processing gparted to mountpoints step tasks."""
+    def qtparted_crashed(self):
+        """qtparted crashed. Ask the user if they want to continue."""
 
-        self.gparted_fstype = {}
+        # TODO cjwatson 2006-07-18: i18n
+        text = ('The advanced partitioner (qtparted) crashed. Further '
+                'information may be found in /var/log/installer/syslog, '
+                'or by running qtparted directly. Do you want to try the '
+                'advanced partitioner again, return to automatic '
+                'partitioning, or quit this installer?')
+        answer = QMessageBox.warning(self.userinterface, 'QTParted crashed',
+                                     text, 'Try again',
+                                     'Automatic partitioning', 'Quit', 0, 0)
+        if answer == 1:
+            self.userinterface.widgetStack.raiseWidget(WIDGET_STACK_STEPS["stepPartDisk"])
+        elif answer == 2:
+            self.current_page = None
+            self.quit()
+        else:
+            self.qtparted_loop()
+
+    def qtparted_to_mountpoints(self):
+        """Processing qtparted to mountpoints step tasks."""
+
+        self.qtparted_fstype = {}
+
+        if self.qtparted_subp is None:
+            self.qtparted_crashed()
+            return
 
         try:
             print >>self.qtparted_subp.stdin, "apply"
@@ -807,19 +847,21 @@ class Wizard:
             self.qtparted_subp.stdin.close()
             self.qtparted_subp.wait()
             self.qtparted_subp = None
+            self.qtparted_crashed()
             return
 
-        # read gparted output of format "- FORMAT /dev/hda2 linux-swap"
-        gparted_reply = self.qtparted_subp.stdout.readline().rstrip('\n')
-        while not gparted_reply.startswith('0 ') and not gparted_reply.startswith('1 '):
-            if gparted_reply.startswith('- '):
-                pre_log('info', 'gparted replied: %s' % gparted_reply)
-                words = gparted_reply[2:].strip().split()
+        # read qtparted output of format "- FORMAT /dev/hda2 linux-swap"
+        qtparted_reply = self.qtparted_subp.stdout.readline().rstrip('\n')
+        while not qtparted_reply.startswith('0 ') and not qtparted_reply.startswith('1 '):
+            pre_log('info', 'qtparted replied: %s' % qtparted_reply)
+            if qtparted_reply.startswith('- '):
+                words = qtparted_reply[2:].strip().split()
                 if words[0].lower() == 'format' and len(words) >= 3:
-                    self.gparted_fstype[words[1]] = words[2]
-            gparted_reply = self.qtparted_subp.stdout.readline().rstrip('\n')
+                    self.qtparted_fstype[words[1]] = words[2]
+            qtparted_reply = self.qtparted_subp.stdout.readline().rstrip('\n')
+        pre_log('info', 'qtparted replied: %s' % qtparted_reply)
 
-        if gparted_reply.startswith('1 '):
+        if qtparted_reply.startswith('1 '):
             # Cancel
             return
 
@@ -827,13 +869,23 @@ class Wizard:
         self.qtparted_subp.stdin.close()
         self.qtparted_subp.wait()
         self.qtparted_subp = None
-        self.qtparted_vbox.remove(self.embed)
+        if self.embed is not None:
+            self.qtparted_vbox.remove(self.embed)
+            del self.embed
+            self.embed = None
 
-        if not gparted_reply.startswith('0 '):
+        if not qtparted_reply.startswith('0 '):
             # something other than OK or Cancel
             return
 
-        self.mountpoint_table = QGridLayout(self.userinterface.mountpoint_frame, 2, 4, 11, 6)
+        children = self.userinterface.mountpoint_frame.children()
+        for child in children:
+            if isinstance(child, QGridLayout):
+                pass
+            else:
+                self.mountpoint_table.remove(child)
+                del child
+
         mountText = "<b>" + get_string("mountpoint_label", self.locale) + "</b>"
         sizeText = "<b>" + get_string("size_label", self.locale) + "</b>"
         partitionText = "<b>" + get_string("device_label", self.locale) + "</b>"
@@ -859,58 +911,62 @@ class Wizard:
             self.part_devices[label] = partition
             self.partition_choices.append(partition)
 
-        # Initialise the mountpoints table.
-        if len(self.mountpoint_widgets) == 0:
-            self.add_mountpoint_table_row()
+        # Reinitialise the mountpoints table.
+        self.mountpoint_widgets = []
+        self.size_widgets = []
+        self.partition_widgets = []
+        self.format_widgets = []
 
-            # Try to get some default mountpoint selections.
-            self.size = get_sizes()
-            selection = get_default_partition_selection(
-                self.size, self.gparted_fstype, self.auto_mountpoints)
+        self.add_mountpoint_table_row()
 
-            # Setting a default partition preselection
-            if len(selection.items()) == 0:
-                self.userinterface.next.setEnabled(False)
-            else:
-                # Setting default preselection values into ComboBox
-                # widgets and setting size values. In addition, next row
-                # is showed if they're validated.
-                for mountpoint, partition in selection.items():
-                    if partition.split('/')[2] not in self.size:
-                        continue
-                    if partition not in self.partition_choices:
-                        # TODO cjwatson 2006-05-27: I don't know why this
-                        # might happen, but it does
-                        # (https://launchpad.net/bugs/46910). Figure out
-                        # why. In the meantime, ignoring this partition is
-                        # better than crashing.
-                        continue
-                    if mountpoint in self.mountpoint_choices:
-                        self.mountpoint_widgets[-1].setCurrentItem(self.mountpoint_choices.index(mountpoint))
-                    else:
-                        self.mountpoint_widgets[-1].setCurrentText(mountpoint)
-                    self.size_widgets[-1].setText(self.set_size_msg(partition))
-                    self.partition_widgets[-1].setCurrentItem(self.partition_choices.index(partition))
-                    if (mountpoint in ('swap', '/', '/usr', '/var', '/boot') or
-                        partition in self.gparted_fstype):
-                        self.format_widgets[-1].setChecked(True)
-                    else:
-                        self.format_widgets[-1].setChecked(False)
-                    if partition not in self.gparted_fstype:
-                        self.format_widgets[-1].setEnabled(True)
-                    if len(get_partitions()) > len(self.partition_widgets):
-                        self.add_mountpoint_table_row()
-                    else:
-                        break
+        # Try to get some default mountpoint selections.
+        self.size = get_sizes()
+        selection = get_default_partition_selection(
+            self.size, self.qtparted_fstype, self.auto_mountpoints)
 
-            # We defer connecting up signals until now to avoid the changed
-            # signal firing while we're busy populating the table.
-            """  Not needed for KDE
-            for mountpoint in self.mountpoint_widgets:
-                self.app.connect(mountpoint, SIGNAL("activated(int)"), self.on_list_changed)
-            for partition in self.partition_widgets:
-                self.app.connect(partition, SIGNAL("activated(int)"), self.on_list_changed)
-            """
+        # Setting a default partition preselection
+        if len(selection.items()) == 0:
+            self.userinterface.next.setEnabled(False)
+        else:
+            # Setting default preselection values into ComboBox widgets and
+            # setting size values. In addition, the next row is shown if
+            # they're validated.
+            for mountpoint, partition in selection.items():
+                if partition.split('/')[2] not in self.size:
+                    continue
+                if partition not in self.partition_choices:
+                    # TODO cjwatson 2006-05-27: I don't know why this might
+                    # happen, but it does
+                    # (https://launchpad.net/bugs/46910). Figure out why. In
+                    # the meantime, ignoring this partition is better than
+                    # crashing.
+                    continue
+                if mountpoint in self.mountpoint_choices:
+                    self.mountpoint_widgets[-1].setCurrentItem(self.mountpoint_choices.index(mountpoint))
+                else:
+                    self.mountpoint_widgets[-1].setCurrentText(mountpoint)
+                self.size_widgets[-1].setText(self.set_size_msg(partition))
+                self.partition_widgets[-1].setCurrentItem(self.partition_choices.index(partition))
+                if (mountpoint in ('swap', '/', '/usr', '/var', '/boot') or
+                    partition in self.qtparted_fstype):
+                    self.format_widgets[-1].setChecked(True)
+                else:
+                    self.format_widgets[-1].setChecked(False)
+                if partition not in self.qtparted_fstype:
+                    self.format_widgets[-1].setEnabled(True)
+                if len(get_partitions()) > len(self.partition_widgets):
+                    self.add_mountpoint_table_row()
+                else:
+                    break
+
+        # We defer connecting up signals until now to avoid the changed
+        # signal firing while we're busy populating the table.
+        """  Not needed for KDE
+        for mountpoint in self.mountpoint_widgets:
+            self.app.connect(mountpoint, SIGNAL("activated(int)"), self.on_list_changed)
+        for partition in self.partition_widgets:
+            self.app.connect(partition, SIGNAL("activated(int)"), self.on_list_changed)
+        """
 
         self.userinterface.mountpoint_error_reason.hide()
         self.userinterface.mountpoint_error_image.hide()
@@ -962,13 +1018,16 @@ class Wizard:
             mountpoint_value = str(self.mountpoint_widgets[i].currentText())
             partition_value = str(self.partition_widgets[i].currentText())
             if partition_value is not None:
-                partition_id = self.part_devices[partition_value]
+                if partition_value in self.part_devices:
+                    partition_id = self.part_devices[partition_value]
+                else:
+                    partition_id = partition_value
             else:
                 partition_id = None
             format_value = self.format_widgets[i].isChecked()
             fstype = None
-            if partition_id in self.gparted_fstype:
-                fstype = self.gparted_fstype[partition_id]
+            if partition_id in self.qtparted_fstype:
+                fstype = self.qtparted_fstype[partition_id]
 
             if mountpoint_value == "":
                 if partition_value in (None, ' '):
@@ -1002,7 +1061,15 @@ class Wizard:
 
         # Processing more validation stuff
         if len(self.mountpoints) > 0:
-            for check in validation.check_mountpoint(self.mountpoints,
+            # Supplement filesystem types from qtparted FORMAT instructions
+            # with those detected from the disk.
+            validate_mountpoints = dict(self.mountpoints)
+            validate_filesystems = get_filesystems(self.qtparted_fstype)
+            for device, (path, format, fstype) in validate_mountpoints.items():
+                if fstype is None and device in validate_filesystems:
+                    validate_mountpoints[device] = \
+                        (path, format, validate_filesystems[device])
+            for check in validation.check_mountpoint(validate_mountpoints,
                                                      self.size):
                 if check == validation.MOUNTPOINT_NOROOT:
                     error_msg.append(get_string(
@@ -1025,6 +1092,22 @@ class Wizard:
                     error_msg.append(get_string(
                         'partman-basicfilesystems/bad_mountpoint',
                         self.locale))
+                elif check == validation.MOUNTPOINT_XFSROOT:
+                    error_msg.append("XFS may not be used on the filesystem "
+                                     "containing /boot. Either use a "
+                                     "different filesystem for / or create a "
+                                     "non-XFS filesystem for /boot.")
+                elif check == validation.MOUNTPOINT_XFSBOOT:
+                    error_msg.append("XFS may not be used on the /boot "
+                                     "filesystem. Use a different filesystem "
+                                     "type for /boot.")
+                elif check == validation.MOUNTPOINT_UNFORMATTED:
+                    error_msg.append("Filesystems used by the system (/, "
+                                     "/boot, /usr, /var) must be reformatted "
+                                     "for use by this installer. Other "
+                                     "filesystems (/home, /media/*, "
+                                     "/usr/local, etc.) may be used without "
+                                     "reformatting.")
 
         # showing warning messages
         self.userinterface.mountpoint_error_reason.setText("\n".join(error_msg))
@@ -1078,11 +1161,14 @@ class Wizard:
                 self.qtparted_subp.stdin.close()
                 self.qtparted_subp.wait()
                 self.qtparted_subp = None
-                self.qtparted_vbox.remove(self.embed)
+                if self.embed is not None:
+                    self.qtparted_vbox.remove(self.embed)
+                    del self.embed
+                    self.embed = None
             self.userinterface.widgetStack.raiseWidget(WIDGET_STACK_STEPS["stepPartDisk"])
             changed_page = True
         elif step == "stepPartMountpoints":
-            self.gparted_loop()
+            self.qtparted_loop()
         elif step == "stepReady":
             self.userinterface.next.setText("Next >")
         if not changed_page:
@@ -1491,7 +1577,7 @@ class Wizard:
 
 
     def set_keyboard_choices(self, choicemap):
-        self.keyboard_choice_map = choicemap
+        self.keyboard_choice_map = dict(choicemap)
         choices = choicemap.keys()
 
         self.userinterface.keyboardlistview.clear()
@@ -1528,8 +1614,11 @@ class Wizard:
         self.userinterface.ready_text.setText(text)
 
     def return_to_autopartitioning (self):
-        """Return from the install progress bar to autopartitioning."""
-        if self.installing:
+        """If the install progress bar is up but still at the partitioning
+        stage, then errors can safely return us to autopartitioning.
+        """
+
+        if self.installing and self.current_page is not None:
             # Go back to the autopartitioner and try again.
             # TODO self.previous_partitioning_page
             #self.live_installer.show()
@@ -1804,6 +1893,8 @@ class MapWidget(QWidget):
             
     def updateCityIndicator(self):
         city = self.getNearestCity(self.width(), self.height(), self.x, self.y)
+        if city is None:
+            return
         self.cityIndicator.setText(city)
         self.cityIndicator.move(self.getPosition(self.cities[city][0], self.cities[city][1], self.width(), self.height()))
         self.cityIndicator.show()
@@ -1812,7 +1903,9 @@ class MapWidget(QWidget):
         pos = mouseEvent.pos()
 
         city = self.getNearestCity(self.width(), self.height(), pos.x(), pos.y());
-        if city == "Edinburgh":
+        if city is None:
+            return
+        elif city == "Edinburgh":
             self.city = "London"
         else:
             self.city = city
